{$reference system.windows.forms.dll}
{$reference System.Drawing.dll}
unit GData;

uses OpenGL,System.Drawing;

const
  EllipseQuality: byte = 255;

var
  WW, WH: word;
  _hdc: HDC;
  MW: real;
  MF: System.windows.forms.Form;

{$region external}

function GetDesktopWindow: system.IntPtr;
external 'User32.dll' name 'GetDesktopWindow';

function GetWindowRect(hWnd: system.IntPtr; var lpRect: system.Drawing.Rectangle): integer;
external 'User32.dll' name 'GetWindowRect';

function GetKeyState(nVirtKey: byte): byte;
external 'User32.dll' name 'GetKeyState';

procedure glGenTextures(n: integer; textures: Pointer);
external 'OpenGL32.dll' name 'glGenTextures';

procedure glBindTexture(target: GLenum; texture: GLuint);
external 'OpenGL32.dll' name 'glBindTexture';

{$endregion}

{$region 2D functions}
function ArcTg(dx, dy: real): real := (dx = 0 ? dy > 0 ? Pi / 2 : -Pi / 2 : ArcTan(dy / dx)) + (dx < 0 ? Pi : 0);

function Rotate(a, b, rot: real): PointF;
begin
  with Result do
  begin
    var r := sqrt(sqr(a) + sqr(b));
    var ang := ArcTg(a, b) + rot;
    X := Cos(ang) * r;
    Y := Sin(ang) * r;
  end;
end;

function Rotate(a: array of PointF; rot: real): array of PointF;
begin
  Result := new PointF[a.Length];
  for i: integer := 0 to a.Length - 1 do
    a[i] := Rotate(a[i].X, a[i].Y, rot);
  Result := a;
end;

function Rotate(p: PointF; rot: real) := Rotate(p.X, p.Y, rot);

procedure ChAngDif(var nCh: Single; tCh, ante: Single);
begin
  var dif := tCh - nCh;
  while dif < -Pi do dif += Pi * 2;
  while dif > Pi do dif -= Pi * 2;
  if abs(dif) < 0.03 then
    nCh := tCh else
    nCh += dif * ante;
end;

function Add(p: array of PointF; dX, dY: real): array of PointF;
begin
  for i: integer := 0 to p.Length - 1 do
  begin
    p[i].X += dX;
    p[i].Y += dY;
  end;
  Result := p;
end;

function Mlt(p: array of PointF; r: real): array of PointF;
begin
  for i: integer := 0 to p.Length - 1 do
  begin
    p[i].X *= r;
    p[i].Y *= r;
  end;
  Result := p;
end;

function Rand(r: real) := (Random * 2 - 1) * r;

{$endregion}

{$region 3DTypes}
type
  
  Point3f = record
    X, Y, Z: Single;
    
    constructor create(X, Y, Z: Single);
    begin
      self.X := X;
      self.Y := Y;
      self.Z := Z;
    end;
  end;
  Point3i = record
    X, Y, Z: integer;
    
    constructor create(X, Y, Z: integer);
    begin
      self.X := X;
      self.Y := Y;
      self.Z := Z;
    end;
  end;

  TexCoord = record
    TX, TY, X, Y, Z: Single;
    
    constructor create(TX, TY, X, Y, Z: Single);
    begin
      self.TX := TX;
      self.TY := TY;
      self.X := X;
      self.Y := Y;
      self.Z := Z;
    end;
  end;
  Texture = class
    
    id: cardinal;
    w, h: word;
    clrs: array of byte;
    
    class function GetPaternBytes(c: integer; func: integer->byte): array of byte;
    begin
      Result := new byte[c];
      for i: integer := 0 to c - 1 do
        Result[i] := func(i);
    end;
    
    procedure Redraw;
    begin
      glBindTexture(GL_TEXTURE_2D, id);
      gluBuild2DMipmaps(GL_TEXTURE_2D, 3, w, h, GL_RGB, GL_UNSIGNED_BYTE, @clrs[0]);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
      glBindTexture(GL_TEXTURE_2D, 0);
    end;
    
    procedure Draw(mode: GLenum; pts: array of TexCoord);
    begin
      glBegin(mode);
      glBindTexture(GL_TEXTURE_2D, id);
      glColor4f(1, 1, 1, 1);
      for i: integer := 0 to pts.Length - 1 do
      begin
        glTexCoord2f(pts[i].TX, pts[i].TY);
        glVertex3f(pts[i].X, pts[i].Y, pts[i].Z);
      end;
      //glBindTexture(GL_TEXTURE_2D, -1);
      glEnd;
    end;
    
    constructor create(w, h: word; clrs: array of byte);
    begin
      glGenTextures(1, @(new cardinal[1](id))[0]);
      self.w := w;
      self.h := h;
      self.clrs := clrs;
    end;
    
    constructor create(re: System.IO.BinaryReader);
    begin
      glGenTextures(1, @id);
      w := re.ReadByte * 256 + re.ReadByte;
      h := re.ReadByte * 256 + re.ReadByte;
      clrs := new byte[w * h * 3];
      re.Read(clrs, 0, clrs.Length);
    end;
    
    constructor create(str: System.IO.Stream);
    begin
      create(new System.IO.BinaryReader(str));
    end;
    
    procedure Save(wr: System.IO.BinaryWriter);
    begin
      wr.Write(w);
      wr.Write(h);
      wr.Write(clrs);
    end;
    
    procedure Save(str: System.IO.Stream) := Save(new System.IO.BinaryWriter(str));
  
  end;

{$endregion}

type
  GraphField = class(System.Windows.Forms.PictureBox)
    timer: System.Windows.Forms.Timer;
    components: System.ComponentModel.IContainer;
    
    RedrawProc: procedure;
    
    procedure TimerTick(sender: Object; e: system.EventArgs) := RedrawProc;
    
    constructor(x, y, w, h, I: integer; Init, Redraw: procedure);
    begin
      self.Left := x;
      self.top := y;
      self.Width := w;
      self.Height := h;
      self.RedrawProc := Redraw;
      components := new System.ComponentModel.Container;
      timer := new System.Windows.Forms.Timer(self.components);
      SuspendLayout;
      timer.Enabled := true;
      timer.Interval := I;
      _hdc := GetDC(self.Handle.ToInt32());
      OpenGLInit(self.Handle);
      
      glEnable(GL_DEPTH_TEST);
      glDepthFunc(GL_LEQUAL);
      glClearDepth(1.0);
      glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
      
      glEnable(GL_BLEND);
      glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
      
      glEnable(GL_TEXTURE_2D);
      Init;
      
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity;
      glViewport(0, 0, Width, Height);
      gluPerspective(45.0, Width / Height, 0.1, real.MaxValue);
      glClearColor(0.0, 0.0, 0.0, 1.0);
      
      glMatrixMode(GL_MODELVIEW);
      
      Cursor.Dispose;
      timer.Tick += TimerTick;
    end;
  end;

type
  gr = record
    
    {$region 2D}
    
    class procedure Polygon(Fill: boolean; cr, cg, cb, ca: Single; pts: array of PointF);
    begin
      if Fill then
        glBegin(GL_POLYGON) else
        glBegin(GL_LINE_LOOP);
      glColor4f(cr, cg, cb, ca);
      for i: integer := 0 to pts.Length - 1 do
        glVertex2f(pts[i].X, pts[i].Y);
      glEnd;
    end;
    
    class procedure Lines(cr, cg, cb, ca: Single; pts: array of PointF);
    begin
      glBegin(GL_LINE_STRIP);
      glColor4f(cr, cg, cb, ca);
      for i: integer := 0 to pts.Length - 1 do
        glVertex2f(pts[i].X, pts[i].Y);
      glEnd;
    end;
    
    class function Ellipse(X, Y, W, H: Single): array of PointF;
    begin
      Result := new PointF[EllipseQuality];
      var dang := Pi * 2 / EllipseQuality;
      for i: integer := 0 to EllipseQuality - 1 do
        Result[i] := new PointF(X + W * Cos(dang * i), Y + H * Sin(dang * i));
    end;
    
    class procedure Rectangle(Fill: boolean; cr, cg, cb, ca: Single; X, Y, W, H: Single);
    begin
      if Fill then
        glBegin(GL_QUADS) else
        glBegin(GL_LINE_LOOP);
      glColor4f(cr, cg, cb, ca);
      glVertex2f(X, Y);
      glVertex2f(X + W, Y);
      glVertex2f(X + W, Y + H);
      glVertex2f(X, Y + H);
      glEnd;
    end;
    
    class procedure Ellipse(Fill: boolean; cr, cg, cb, ca: Single; X, Y, W, H: Single);
    begin
      Polygon(Fill, cr, cg, cb, ca, Ellipse(X, Y, W, H));
    end;
    
    {$endregion}
    
    {$region 3D}
    
    class procedure Polygon(Fill: boolean; cr, cg, cb, ca: Single; pts: array of Point3f);
    begin
      if Fill then
        glBegin(GL_POLYGON) else
        glBegin(GL_LINE_LOOP);
      glColor4f(cr, cg, cb, ca);
      for i: integer := 0 to pts.Length - 1 do
        glVertex3f(pts[i].X, pts[i].Y, pts[i].Z);
      glEnd;
    end;
    
    class procedure Sphere(cr, cg, cb, ca, X, Y, Z, dX, dY, dZ: Single; pc: integer);
    begin
      Randomize(0);
      glColor4f(cr, cg, cb, ca);
      var pts := new Point3f[pc * 2, pc + 1];
      var dangx := Pi / pc;
      var dangy := Pi / pc;
      for ix: integer := 0 to pc * 2 - 1 do
        for iy: integer := 0 to pc do
        begin
          var AX := ix * dangx;
          var AY := iy * dangy;
          pts[ix, iy] := new Point3f(X + dX * Cos(AX) * Sin(AY), Y + dY * Cos(AY), Z + dZ * Sin(AX) * Sin(AY));
        end;
      
      
      glBegin(GL_QUADS);
      var lx := pc * 2 - 1;
      for ix: integer := 0 to pc * 2 - 1 do
      begin
        for iy: integer := 0 to pc - 1 do
        begin
          with pts[ix + 0, iy + 0] do glVertex3f(X, Y, Z);
          with pts[lx + 0, iy + 0] do glVertex3f(X, Y, Z);
          with pts[lx + 0, iy + 1] do glVertex3f(X, Y, Z);
          with pts[ix + 0, iy + 1] do glVertex3f(X, Y, Z);
        end;
        lx := ix;
      end;
      glEnd;
    end;
    
    class procedure Lines(cr, cg, cb, ca: Single; pts: array of Point3f);
    begin
      glBegin(GL_LINE_STRIP);
      glColor4f(cr, cg, cb, ca);
      for i: integer := 0 to pts.Length - 1 do
        glVertex3f(pts[i].X, pts[i].Y, pts[i].Z);
      glEnd;
    end;
  
    {$endregion}
  
  end;

begin
  begin
    var a: system.Drawing.Rectangle;
    GetWindowRect(GetDesktopWindow, a);
    WW := a.Width;
    WH := a.Height;
  end;
end.